{ 

 (* Grail parser.  Adapted from example in Moscow ML distribution. *)

 open Lexing Parser;

 exception LexicalError of string * int * int (* (message, loc1, loc2) *)

 fun lexerError lexbuf s = 
     raise LexicalError (s, getLexemeStart lexbuf, getLexemeEnd lexbuf);

 val commentStart = ref 0;  (* Start of outermost comment being scanned *)
 
 fun commentNotClosed lexbuf =
     raise LexicalError ("Comment not terminated", 
                         !commentStart, getLexemeEnd lexbuf);
     
 val commentDepth = ref 0;  (* Current comment nesting *)


 (* Scan keywords as identifiers and use this function to distinguish them. *)
 (* If the set of keywords is large, use an auxiliary hashtable.            *)

 fun keyword s =
     case s of
          "alias" => ALIAS
	| "class" => CLASS
	| "field" => FIELD
	| "method" => METHOD
	| "let" => LET
	| "fun" => FUN
	| "in" => IN
	| "and" => AND
	| "end" => END
	| "val" => VAL
	| "if" => IF
	| "then" => THEN
	| "else" => ELSE
	| "new" => NEW
	| "checkcast" => CHECKCAST
	| "instanceof" => INSTANCEOF
	| "invokestatic" => INVOKESTATIC
	| "invokevirtual" => INVOKEVIRTUAL
	| "invokespecial" => INVOKESPECIAL
	| "getfield" => GETFIELD
	| "putfield" => PUTFIELD
	| "getstatic" => GETSTATIC
	| "putstatic" => PUTSTATIC
	| "add" => ADD
	| "mul" => MUL
	| "sub" => SUB
	| "div" => DIV
	| "mod" => MOD
	| "null" => NULL
	| "int" => INT
        | "boolean" => BOOLEAN
        | "float" => FLOAT
	| "ref" => REF
	| "void" => VOID
        | "public" => PUBLIC
        | "private" => PRIVATE
        | "protected" => PROTECTED
        | "static" => STATIC
        | "final" => FINAL
        | "make" => MAKE
        | "get" => GET
        | "set" => SET
        | "length" => LENGTH
        | "empty" => EMPTY
        | "itof" => ITOF
        | "ftoi" => FTOI
        | "string" => STRING
        | _ => stringval s;
}

rule Token = parse
    [` ` `\t` `\n` `\r`]     { Token lexbuf }
  | `-`?[`0`-`9`]+      { case Int.fromString (getLexeme lexbuf) of
                               NONE   => lexerError lexbuf "internal error"
			    | SOME i => intval i
			}
  | `-`?[`0`-`9`]+`.`[`0`-`9`]+([`e``E`]`-`?[`0`-`9`]+)?
      { case Real.fromString (getLexeme lexbuf) of
                               NONE   => lexerError lexbuf "internal error"
			    | SOME i => floatval i
                        }  

  | [`_``a`-`z``A`-`Z``$``?`][`a`-`z``A`-`Z``0`-`9``_``.``:``#``$`]*(`<`[`a`-`z`]+`>`)?`'`*
                        { keyword (getLexeme lexbuf) }
  | `<`[`a`-`z`]+`>` { stringval (getLexeme lexbuf) }
(*  | `"`([^`\\``"`]*|"\\\\"|`\\``"`)`"` { quoteval (getLexeme lexbuf)  } weirdness *)
(*  | `"`([^`\\``"`]*|"\\\\"|"\\n")`"` { quoteval (getLexeme lexbuf)  }*)
  | `"`[^`"`]*`"` { quoteval (getLexeme lexbuf)  } 

  | "/*"                { commentStart := getLexemeStart lexbuf;
                          commentDepth := 1; 
                          SkipComment lexbuf; Token lexbuf }
  | "//"                { DiscardRestOfLine lexbuf; Token lexbuf }
  | `=`                 { EQUALS }
  | "<>"                { NOTEQUAL }
  | `<`                 { LESS }
  | "<="                { LESSEQ }
  | `>`                 { GREATER }
  | ">="                { GREATEREQ }
  | `(`                 { LPAREN }
  | `)`                 { RPAREN }
  | `{`                 { LBRACE }
  | `}`                 { RBRACE }
  | `[`                 { LSQ }
  | `]`                 { RSQ }
  | `,`                 { COMMA }
  | eof                 { EOF }
  | _                   { lexerError lexbuf "Illegal symbol in input" }

and SkipComment = parse
     "*/"               { commentDepth := !commentDepth - 1;  
                          if !commentDepth = 0 then ()
                          else SkipComment lexbuf 
                        } 
   | "/*"               { commentDepth := !commentDepth + 1; 
                          SkipComment lexbuf }
   | (eof | `\^Z`)      { commentNotClosed lexbuf }
   | _                  { SkipComment lexbuf }

and DiscardRestOfLine = parse 
     (eof | `\n`) { () }
   | _ { DiscardRestOfLine lexbuf }

;
