(* Constpool.sml
 *
 * Peter Bertelsen
 * December 1997
 *)

exception InvalidEntry of string

fun entryError s = raise InvalidEntry("Constpool." ^ s)

datatype index = IDX of int

val maxPool = 0xffff

fun makeIndex k = if 0 <= k andalso k < maxPool then IDX k
		  else raise Overflow

fun indexValue (IDX i) = i

(* NOTE: index values produced by the insert functions will never
 * exceed maxPool-1, due to the use of Array.sub in insert.
 *)

datatype entry =
    CPutf8 of string
  | CPint of Int32.int
  | CPfloat of Word8Vector.vector
  | CPlong of Int64.int
  | CPdouble of Word8Vector.vector
  | CPclass of index
  | CPstring of index
  | CPfieldref of {class: index, nameType: index}
  | CPmethodref of {class: index, nameType: index}
  | CPimethodref of {class: index, nameType: index}
  | CPnametype of {name: index, desc: index}
  | CPunused  (* placeholder *)

(* NOTE: float and double entries are stored as byte vectors, rather
 * than Real32.real/Real64.real values, since polymorphic equality is
 * used in the constant pool implementation.
 *)

datatype pool =
    CP of {pool  : entry Array.array ref,
	   cache : (entry, index) Polyhash.hash_table,
	   count : int ref}

(* NOTE: the first valid entry in the constant pool is pool[1]. *)

val exnFind = Bytecode.InternalError
              "Constpool.insert: unexpected use of Polyhash.find"

fun create () =
    CP {pool  = ref (Array.array(maxPool, CPunused)),
	cache = Polyhash.mkPolyTable(1021, exnFind),
	count = ref 1}

fun insert size (CP {pool, cache, count}) entry =
    let val count' = !count
	val index  = IDX count'
    in
	case Polyhash.peekInsert cache (entry, index) of
	    NONE => (Array.update(!pool, count', entry);
		     count:= count' + size;
		     index)
	  | SOME index' => index'
    end

val insert1 = insert 1   (* insert constant occupying one entry *)
val insert2 = insert 2   (* insert constant occupying two entries *)

fun insUtf8 cp str = insert1 cp (CPutf8 str)

fun insInt cp k = insert1 cp (CPint k)

fun insFloat cp r = insert1 cp (CPfloat (Real32.toBytes r))

fun insLong cp k = insert2 cp (CPlong k)

fun insDouble cp r = insert2 cp (CPdouble (Real64.toBytes r))

fun insString cp str =
    let val ins = insert1 cp
    in
	ins(CPstring (ins(CPutf8 str)))
    end

fun insClass' ins class =
    let val name = Jvmtype.qualName class
    in
	ins(CPclass (ins(CPutf8 name)))
    end

fun insClass cp = insClass'(insert1 cp)

fun insArrayClass cp t =
    let open Jvmtype
	val ins = insert1 cp
	val desc = (case t of
			Tarray t'    => typeDesc t
		      | Tclass class => qualName class
		      | _ => entryError "insArrayClass: invalid array class")
    in
	ins(CPclass (ins(CPutf8 desc)))
    end

fun insNameType' ins name desc = ins(CPnametype {name = ins(CPutf8 name),
						 desc = ins(CPutf8 desc)})

fun insNameType cp {name, desc} = insNameType' (insert1 cp) name desc

fun insFieldref cp {class, name, ty} =
    let val ins  = insert1 cp
	val desc = Jvmtype.typeDesc ty
    in
	ins(CPfieldref {class    = insClass' ins class,
			nameType = insNameType' ins name desc})
    end

fun insMethodref cp {class, name, msig} =
    let val ins  = insert1 cp
	val desc = Jvmtype.methodDesc msig
    in
	ins(CPmethodref {class    = insClass' ins class,
			 nameType = insNameType' ins name desc})
    end

fun insIMethodref cp {class, name, msig} =
    let val ins = insert1 cp
	val desc = Jvmtype.methodDesc msig
    in
	ins(CPimethodref {class    = insClass' ins class,
			  nameType = insNameType' ins name desc})
    end

local
    open Bytecode
in
    fun insConst cp value =
	(case value of
	     Cint i    => insInt cp i
	   | Cfloat f  => insFloat cp f
	   | Clong l   => insLong cp l
	   | Cdouble d => insDouble cp d
	   | Cstring s => insString cp s
	   | _ => entryError "insConst: invalid constant")
end

fun lookup (CP {pool, count, ...}) (IDX i) =
    if 1 <= i andalso i < !count then 
	Array.sub(!pool, i)
    else
	raise Subscript

fun charToUtf8 (c, res) =  (* convert character to UTF-8 representation *)
    let val k = ord c
    in
	if 1 <= k andalso k <= 127 then
	    (Word8.fromInt k) :: res
	else
	 (* if k <= 2047 then *)
		let
		    val x = 0xc0 + (k div 0x40)
		    val y = 0x80 + (k mod 0x40)
		in
		    (Word8.fromInt x) :: (Word8.fromInt y) :: res
		end
	 (* else
		let
		    val k' = k div 0x40
		    val x  = 0xe0 + (k' div 0x40)
		    val y  = 0x80 + (k' mod 0x40)
		    val z  = 0x80 + (k  mod 0x40)
		in
		    (Word8.fromInt x) :: (Word8.fromInt y) ::
		    (Word8.fromInt z) :: res
		end *)
         (* NOTE: the latter section will be necessary if WideChar.char is
	  * used for characters
	  *)
    end
    
fun emit out (CP {pool, count, ...}) =
    let val emitInt  = Int32.emit out
	val emitLong = Int64.emit out
	val emitU2   = Word16.emit out
	val emitU2i  = emitU2 o Word16.fromInt

	fun emitIndex (IDX i) = emitU2i i

	fun emit' (CPutf8 s) =
	        let val cs = CharVector.foldr charToUtf8 [] s
		in
		    out 0w1;   (* CONSTANT_Utf8 *)
		    emitU2i(length cs);
		    List.app out cs
		end
	  | emit' (CPint i) =
		(out 0w3;   (* CONSTANT_Integer *)
		 emitInt i)
	  | emit' (CPfloat bs) =
		(out 0w4;   (* CONSTANT_Float *)
		 Word8Vector.app out bs)
	  | emit' (CPlong l) =
		(out 0w5;   (* CONSTANT_Long *)
		 emitLong l)
	  | emit' (CPdouble bs) =
		(out 0w6;   (* CONSTANT_Double *)
		 Word8Vector.app out bs)
	  | emit' (CPclass i) =
		(out 0w7;   (* CONSTANT_Class *)
		 emitIndex i)
	  | emit' (CPstring i) =
		(out 0w8;   (* CONSTANT_String *)
		 emitIndex i)
	  | emit' (CPfieldref {class = c, nameType = nt}) =
		(out 0w9;   (* CONSTANT_Fieldref *)
		 emitIndex c;
		 emitIndex nt)
	  | emit' (CPmethodref {class = c, nameType = nt}) =
		(out 0w10;   (* CONSTANT_Methodref *)
		 emitIndex c;
		 emitIndex nt)
	  | emit' (CPimethodref {class = c, nameType = nt}) =
		(out 0w11;   (* CONSTANT_IMethodref *)
		 emitIndex c;
		 emitIndex nt)
	  | emit' (CPnametype {name = n, desc = d}) =
		(out 0w12;   (* CONSTANT_NameAndType *)
		 emitIndex n;
		 emitIndex d)
          | emit' CPunused = ()
    in
	emitU2i(!count);
	Array.appi (fn (_, e) => emit' e) (!pool, 1, SOME(!count))
    end
