(* Grail abstract syntax *)

structure GrailAbsyn :> sig

datatype ClassDef = CDEF of Classdecl.access_flag list * string * 
    FieldDef list * MethodDef list
and FieldDef = FDEF of Classdecl.access_flag list * Ty * string
and MethodDef = MDEF of Classdecl.access_flag list * RTy * string * 
    (Ty*Var) list * MethodBody
and MethodBody = MBODY of LetDec list * FunDec list * Result
and LetDec = 
    VALdec of string * PrimOp
  | VOIDdec of PrimOp
and FunDec = FDEC of string * (Ty*Var) list * FunBody
and FunBody = FUNbody of LetDec list * Result
and Result =
    PRIMres of PrimRes
  | CHOICEres of Value * Test * Value * PrimRes * PrimRes
and PrimRes = 
    OPres of PrimOp
  | VOIDres
  | FUNres of string * Var list
and PrimOp =
    VALop of Value
  | BINop of BinOp * Value * Value
  | NEWop of MethDesc * Value list
  | CHECKCASTop of string * Var
  | INSTANCEop of string * Var
  | INVOKESTATICop of MethDesc * Value list
  | INVOKEVIRTUALop of Var * MethDesc * Value list
  | INVOKESPECIALop of Var * MethDesc * Value list
  | GETFIELDop of Var * FieldDesc
  | PUTFIELDop of Var * FieldDesc * Value
  | GETSTATICop of FieldDesc
  | PUTSTATICop of FieldDesc * Value
  | MAKEop of Value * Value
  | GETop of Value * Value
  | SETop of Value * Value * Value
  | LENGTHop of Value  
  | EMPTYop of Value * Ty
  | FTOIop of Value
  | ITOFop of Value
and MethDesc = MDESC of RTy * string * Ty list
and FieldDesc = FDESC of Ty * string
and Test = EQtest | NEtest | Ltest | LEtest | Gtest | GEtest 
and BinOp = ADDop | SUBop | MULop | DIVop | MODop
and Value = 
    VARval of Var
  | INTval of int
  | FLOATval of real
  | STRINGval of string
  | NULLval of string
and Ty = (* many of these aren't accessible from Grail *)
    BYTEty 
  | CHARty 
  | DOUBLEty 
  | FLOATty
  | INTty 
  | LONGty 
  | SHORTty 
  | BOOLEANty 
  | REFty of string
  | ARRAYty of Ty
withtype Var = string
and RTy = Ty option
;

val prClassDef: ClassDef -> unit

end

= struct

datatype ClassDef = CDEF of Classdecl.access_flag list * string * 
    FieldDef list * MethodDef list
and FieldDef = FDEF of Classdecl.access_flag list * Ty * string
and MethodDef = MDEF of Classdecl.access_flag list * RTy * string * 
    (Ty*Var) list * MethodBody
and MethodBody = MBODY of LetDec list * FunDec list * Result
and LetDec = 
    VALdec of string * PrimOp
  | VOIDdec of PrimOp
and FunDec = FDEC of string * (Ty*Var) list * FunBody
and FunBody = FUNbody of LetDec list * Result
and Result =
    PRIMres of PrimRes
  | CHOICEres of Value * Test * Value * PrimRes * PrimRes
and PrimRes = 
    OPres of PrimOp
  | VOIDres
  | FUNres of string * Var list
and PrimOp =
    VALop of Value
  | BINop of BinOp * Value * Value
  | NEWop of MethDesc * Value list
  | CHECKCASTop of string * Var
  | INSTANCEop of string * Var
  | INVOKESTATICop of MethDesc * Value list
  | INVOKEVIRTUALop of Var * MethDesc * Value list
  | INVOKESPECIALop of Var * MethDesc * Value list
  | GETFIELDop of Var * FieldDesc
  | PUTFIELDop of Var * FieldDesc * Value
  | GETSTATICop of FieldDesc
  | PUTSTATICop of FieldDesc * Value
  | MAKEop of Value * Value
  | GETop of Value * Value
  | SETop of Value * Value * Value
  | LENGTHop of Value  
  | EMPTYop of Value * Ty
  | FTOIop of Value
  | ITOFop of Value
and MethDesc = MDESC of RTy * string * Ty list
and FieldDesc = FDESC of Ty * string
and Test = EQtest | NEtest | Ltest | LEtest | Gtest | GEtest 
and BinOp = ADDop | SUBop | MULop | DIVop | MODop
and Value = 
    VARval of Var
  | INTval of int
  | FLOATval of real
  | STRINGval of string
  | NULLval of string
and Ty = (* many of these aren't accessible from Grail *)
    BYTEty 
  | CHARty 
  | DOUBLEty 
  | FLOATty
  | INTty 
  | LONGty 
  | SHORTty 
  | BOOLEANty 
  | REFty of string
  | ARRAYty of Ty
withtype Var = string
and RTy = Ty option
;

exception gfError of string;

(************ Printing abstract syntax ************)

val indent = ref "";
fun incIndent() = indent := (!indent) ^ "   ";
fun decIndent() = indent := String.extract(!indent, 3, NONE);

val Pr = TextIO.print;
fun Prl s = TextIO.print (s^"\n");
fun nl(s) =  TextIO.print ("\n" ^ (!indent) ^ s);

fun tyToString BYTEty       = "byte"
  | tyToString CHARty       = "char"
  | tyToString DOUBLEty     = "double"
  | tyToString FLOATty      = "float"
  | tyToString INTty        = "int"
  | tyToString LONGty       = "long"
  | tyToString SHORTty      = "short"
  | tyToString BOOLEANty    = "boolean"
  | tyToString (ARRAYty(t)) = tyToString(t) ^ "[]"
  | tyToString (REFty(s))   = s

fun rtyToString (NONE) = "void"
  | rtyToString (SOME ty) = tyToString ty

fun valToString (VARval v) = v
  | valToString (INTval n) = 
    let
	val s = Int.toString (Int.abs n)
    in
	if (n<0) then "-" ^ s else s
    end
  | valToString (FLOATval x) =
    let (* There's a bug in MLj's Real.toString anyway *)
	val s = Real.toString (Real.abs x)
    in
	if (x<0.0) then "-" ^ s else s
    end
  | valToString (STRINGval s) = "\"" ^ s ^ "\""
  | valToString (NULLval(s)) = "null[" ^ s ^ "]"


fun listToString toString l = 
    let 
	fun f [] = ""
	  | f [x] = toString x
	  | f (h::t) = (toString h) ^ ", " ^ f t
    in
	"(" ^ f l ^ ")"
    end
;

val strListToString = listToString (fn x => x)

fun prList toString l = 
       Pr (listToString toString l)

val prVarList = prList (fn (s:string) => s)

val prIntList = prList Int.toString

val prTyList = prList tyToString

val prTyVarList = prList (fn (t,v) => (tyToString t ^ " " ^ v))

val prValList = prList valToString

fun nth([],_) = raise gfError "nth: list to short"
  | nth (h::_,0) = h
  | nth (_::t,n) = nth(t,n-1)

fun prClassDef (CDEF (flags, cname, fdefs, mdefs)) =
(
    app prFlag flags;
    Pr ("class " ^ cname ^ " {");
    incIndent();
    app prFieldDef fdefs;
    app prMethodDef mdefs;
    decIndent();
    Prl "}"
)

and prFieldDef (FDEF(flags, ty, name )) =
    (
     nl "field ";
     app prFlag flags;
     prTy ty;
     Pr " ";
     Prl name
)

and prMethodDef (MDEF(flags, rty, name, params, mbody)) =
    (
     nl "method ";
     app prFlag flags;
     prRTy rty;
     Pr (" " ^ name ^ " ");
     prTyVarList params;
     Pr " =";
     prMethodBody mbody
     )

and prMethodBody (MBODY(letDecs, funDecs, result)) =
    (
     nl "let";
     incIndent();
     app prLetDec letDecs;
     app prFunDec funDecs;
     decIndent();
     nl "in";
     prResult result;
     nl "end\n"
     )

and prLetDec (VALdec(id, p)) = 
    ( nl("val " ^ id ^ " = "); prPrimOp p )
  | prLetDec (VOIDdec p) = 
    ( nl "val () = "; prPrimOp p)

and prFunDec (FDEC (fname, params, fbody)) =
    (
     TextIO.print "\n";
     nl("fun " ^ fname ); 
     prTyVarList params; 
     Pr " =";
     prFunBody fbody
)

and prFunDecList [] = ()  (* just to see if we can get round the bug *)
  | prFunDecList (h::t) = (prFunDec h; prFunDecList t)

and prFunBody (FUNbody(letdecs, result)) =
    if letdecs = nil then prResult result
    else
	(nl "let";
	 incIndent();
	 app prLetDec letdecs;
	 decIndent();
	 nl "in";
	 prResult result;
	 nl "end"
	 )

and prResult (PRIMres p) = (incIndent(); nl ""; prPrimRes p; decIndent())
  | prResult (CHOICEres (v, tst, w, p1, p2)) =
    ( 
     incIndent();
     nl "if "; prValue v; prTest tst; prValue w;
     nl "then "; prPrimRes p1;
     nl "else "; prPrimRes p2;
     decIndent()
     )

and prPrimRes (OPres p) = prPrimOp p 
  | prPrimRes (VOIDres) = Pr "()"
  | prPrimRes (FUNres (name, args)) = (Pr name; prVarList args)
		
and prPrimOp (VALop v) = prValue v
  | prPrimOp (BINop(b, v, w)) = (prBinOp b; prValue v; Pr " "; prValue w)
  | prPrimOp (NEWop(m,v)) = prNew (m, v)
  | prPrimOp (CHECKCASTop(s,v)) = Prl ("checkcast " ^ s ^ " " ^ v )
  | prPrimOp (INSTANCEop(s,v)) =  Prl ("instance " ^ s ^ " " ^ v )
  | prPrimOp (INVOKESTATICop(mdesc, args)) =
    ( Pr "invokestatic "; prMethDesc mdesc; prValList args )
  | prPrimOp (INVOKEVIRTUALop(v, mdesc, args)) =
    ( Pr ("invokevirtual " ^ v ^ " "); prMethDesc mdesc; prValList args )
  | prPrimOp (INVOKESPECIALop(v, mdesc, args)) =
    ( Pr ("invokespecial " ^ v ^ " "); prMethDesc mdesc; prValList args )
  | prPrimOp (GETFIELDop(v, fdesc)) =
    ( Pr ("getfield " ^ v ^ " " ); prFieldDesc fdesc )
  | prPrimOp (PUTFIELDop(v, fdesc, w)) =
    ( Pr ("putfield " ^ v ^ " "); prFieldDesc fdesc; Pr " "; prValue w )
  | prPrimOp (GETSTATICop(fdesc)) = 
    ( Pr "getstatic "; prFieldDesc fdesc )
  | prPrimOp (PUTSTATICop(fdesc,w)) =
    ( Pr "putstatic "; prFieldDesc fdesc; Pr " "; prValue w)
  | prPrimOp (MAKEop(a, i)) = 
    ( Pr "make "; prValue a; Pr " "; prValue i)
  | prPrimOp (SETop(aa, i, a)) =
    ( Pr "set "; prValue aa; Pr " "; prValue i; Pr " "; prValue a)
  | prPrimOp (LENGTHop(aa)) =
    ( Pr "length "; prValue aa)
  | prPrimOp (EMPTYop(a,ARRAYty(ty))) =
    ( Pr "empty "; prValue a; Pr " "; prTy ty)
  | prPrimOp (GETop(a,i)) = 
    ( Pr "get "; prValue a; Pr " "; prValue i)
  | prPrimOp (FTOIop(f)) = 
    ( Pr "ftoi "; prValue f)
  | prPrimOp (ITOFop(i)) = 
    ( Pr "itof "; prValue i)
  | prPrimOp _ =
    ( raise gfError "Not implemented yet" )
    
(* Stupid structure due to compiler bug *)
and prNew (mdesc, vals) = 
    (case mdesc of 
	 MDESC(rty, name, types) => (Pr ("new <" ^ name);
				     prTyList types;
				     Pr "> ";
				     prValList vals
				     )
       | _ => raise gfError "error printing new"
     )
and prMethDesc (MDESC(rty, name, types)) =
    (Pr  "<";
     prRTy rty;
     Pr (" " ^ name ^ " ");
     prTyList types;
     Pr "> "
     )
and prFieldDesc (FDESC(ty, name)) = 
    (Pr "<";
     prTy ty;
     Pr " ";
     Pr name;
     Pr ">"     
     )

and prTest EQtest = Pr " = "
  | prTest NEtest = Pr " <> "
  | prTest Ltest = Pr " < "
  | prTest LEtest = Pr " <= "
  | prTest Gtest = Pr " > "
  | prTest GEtest = Pr " >= "

and prBinOp ADDop = Pr "add "
  | prBinOp SUBop = Pr "sub "
  | prBinOp MULop = Pr "mul "
  | prBinOp DIVop = Pr "div "
  | prBinOp MODop = Pr "mod "

and prValue v = Pr (valToString v)

and prTy t = Pr (tyToString t)

and prRTy r = Pr (rtyToString r)

and prFlag Classdecl.ACCpublic =    Pr "public "
  | prFlag Classdecl.ACCprivate =   Pr "private "
  | prFlag Classdecl.ACCprotected = Pr "protected "
  | prFlag Classdecl.ACCstatic =    Pr "static "
  | prFlag Classdecl.ACCfinal =     Pr "final "
  | prFlag Classdecl.ACCvolatile =  Pr "volatile "
  | prFlag Classdecl.ACCtransient = Pr "transient "
  | prFlag Classdecl.ACCsynchronized = Pr "synchronized "
  | prFlag Classdecl.ACCnative =    Pr "native "
  | prFlag Classdecl.ACCabstract =  Pr "abstract "
  | prFlag Classdecl.ACCsuper =     Pr ""
  | prFlag Classdecl.ACCinterface = Pr "interface " 
;

end
