(*======================================================================*)
(* gf: Grail decompiler                					*)
(*======================================================================*)
structure gf :> 
sig
  val main : string option array option -> unit
end 
=
struct

val nth = List.nth;
val length = List.length;
val take = List.take;
val drop = List.drop;
val reverse = List.rev;
val tabulate = List.tabulate;
val filter = List.filter
val foldl = List.foldl
(* MLj doesn't like it if we try to refer to these directly *)

fun snd (_,y) = y

open org.apache.bcel.classfile;
open org.apache.bcel.generic;
open java.io;
open java.util;
open org.apache.bcel.Constants;
open org.apache.bcel.Repository;
open Classdecl;

exception gfError of string

open GrailAbsyn
open Bytecode
open printJvmInstr

val varInfo = ref ([]: (int*string*Ty) list)

(********** Decompilation to Grail abstract syntax **********)

fun arrayToList a =
let fun g a j l =
    if j < 0 then l
    else g a (j-1) ((valOf(Array.sub(a,j)))::l)

in
    g a (Array.length(a)-1) []
end


fun getFlags (m: AccessFlags) = 
let
    val stuff = 
    [
     (m.#isInterface(), ACCinterface),
     (m.#isNative(), ACCnative),
     (m.#isVolatile(), ACCvolatile),
     (m.#isTransient(), ACCtransient),
     (m.#isSynchronized(), ACCsynchronized),
     (m.#isAbstract(), ACCabstract),
     (m.#isFinal(), ACCfinal),
     (m.#isStatic(), ACCstatic),
     (m.#isProtected(), ACCprotected),
     (m.#isPrivate(), ACCprivate),
     (m.#isPublic(), ACCpublic)
     ]
    fun get nil acc = acc
      | get ((b,d)::t) acc = if b then get t (d::acc) else get t acc
in 
    get stuff []
end


fun jTyDescToTy s =
let 
in
    if String.extract (s, String.size(s)-1, NONE) = "]" 
	then ARRAYty (jTyDescToTy (String.substring(s,0,String.size(s)-2)))
    else case s of
	"byte"    => BYTEty
      | "char"    => CHARty
      | "double"  => DOUBLEty
      | "float"   => FLOATty
      | "int"     => INTty
      | "long"    => LONGty
      | "short"   => SHORTty
      | "boolean" => BOOLEANty
      | "void" => raise gfError "Unexpected void"
      | _ => REFty s
end

(* From GrailAbsyn. Should remove *)
fun tyToString BYTEty       = "byte"
  | tyToString CHARty       = "char"
  | tyToString DOUBLEty     = "double"
  | tyToString FLOATty      = "float"
  | tyToString INTty        = "int"
  | tyToString LONGty       = "long"
  | tyToString SHORTty      = "short"
  | tyToString BOOLEANty    = "boolean"
  | tyToString (ARRAYty(t)) = tyToString(t) ^ "[]"
  | tyToString (REFty(s))   = s

local
    val name = ref #"a"
in
    fun resetNames() = name := #"a"
    fun newName() = 
	let 
	    val n = String.str (!name)
	    val () = name := Char.succ (!name)
	in
	    n
	end
end

fun jfieldToFieldDef (f: Field)  =
let
    val flags = getFlags (f :> AccessFlags)
    val sign = valOf (f.#getSignature())
    val sign2 = valOf(Utility.signatureToString(sign, false))
in
    FDEF(flags, jTyDescToTy sign2, valOf(f.#getName()))
end



fun prAttrs [] = () 
  | prAttrs ((h:Attribute)::t) = 
    let 
	val x = valOf(h.#toString())
    in
	(* TextIO.print(x^"\n"); *)
	prAttrs t
    end

fun prInsts [] = () 
  | prInsts ((h:Instruction)::t) = 
    let 
	val x = valOf(h.#toString())
    in
	(* TextIO.print(x^"\n"); *)
	prInsts t
    end


datatype gLabel = 
    FUNlabel of Label.label 
  | IFlabel of Label.label

(* We need to keep track of which labels are targets of goto instructions
   since these are the labels corresponding to the start of Grail functions,
   and we want to know about these when decoding the metadata *)

fun lookup  (v: InstructionHandle) ((ref l):  (InstructionHandle * gLabel) list ref ) = 
    let fun
	lookup' (v: InstructionHandle) [] = NONE
	  | lookup' (v: InstructionHandle) ((w: InstructionHandle, p)::tail) =
		if v.#equals(w) then SOME p else lookup' v tail
(*	val _ = TextIO.print ("Looking up " ^ valOf(v.#toString()) ^ "\n")*)
    in 
	lookup' v l
    end
;

fun find (v: InstructionHandle) (h :  (InstructionHandle * gLabel) list ref) = 
    case lookup v h of 
	NONE => raise gfError ("Can't find target")
      | SOME p => p
;


local 
    val nextlab = ref Label.freshLabels : Label.labels ref
    val (_,cs) = Label.newLabel(!nextlab)   (* Horrible *)
    val (_,ce) = Label.newLabel(!nextlab)   (* Horrible *)
    val codeLabels: (InstructionHandle * gLabel) list ref 
	= ref [];

    val funLabels: Label.label list ref = ref [];
in      
    val codeStart = ref cs;
    val codeEnd = ref cs;  (* Both of these are initial values which will never be used *)

    fun newLabel () = (* We need this to be globally visible so that 
		         we can generate new labels for "if ..." *)
        let val (labels, lab) = Label.newLabel (!nextlab)
        in nextlab := labels; lab
	end

    fun resetLabels () = 
	(nextlab := Label.freshLabels; 
	 codeStart := newLabel();
	 codeEnd := newLabel();
	 codeLabels := [])
	
    fun lookupLabel (f:InstructionHandle) = lookup f codeLabels;
    fun findLabel   (f:InstructionHandle) = find f codeLabels;
	
    fun setFunLabels() =
	let
	    fun prune [] = []
	      | prune ((_,IFlabel _)::t) = prune t
	      | prune ((_,FUNlabel l)::t) = l::(prune t)
	in
	    funLabels := reverse (prune (!codeLabels))
	end

    fun getFunLabels() = !funLabels
		    
    fun saveLabel (NONE, _) = raise gfError "Trying to declare null target"
      | saveLabel (SOME ih, isGoto) = 
	case lookupLabel ih of 
	    SOME _ => () (*TextIO.print ("Redeclaring target\n\n")*)
	(* Should maybe check here that jump is of same type as 
	   during declaration *)
	  | NONE => let 
			val loc = newLabel() 
			val loc' = if isGoto then FUNlabel loc else IFlabel loc
(*			val _ = TextIO.print ("Declaring label " 
					      ^ Label.toString(loc)) 
			val _ = if isGoto then TextIO.print " (fun)\n"
				else TextIO.print " (if)\n"
			val _ = TextIO.print ("{ " ^ valOf(ih.#toString(true)) ^ " }\n\n")
*)
		    in 
			codeLabels := (ih, loc')::(!codeLabels) 
		    end

end;

fun getTargets [] = ()
    | getTargets ((h: InstructionHandle)::t) = 
      (if _instanceof BranchHandle h
	  then 
	      let val bi = (valOf(h.#getInstruction())):>BranchInstruction
	      in if _instanceof Select bi then raise gfError ("Found switch instruction")
		 else
		     if _instanceof GotoInstruction bi then saveLabel (bi.#getTarget(),true)
		     else saveLabel (bi.#getTarget(), false)
	      end
      else (); getTargets t )
		       


fun qnameToJClass name = 
    if name ="" then Jvmtype.class{pkgs = [], name = ""} else 
    let val l = String.tokens (fn c => (c = #".")) name
	val n = length l
	val pkgs = take (l,n-1)
	val name' = nth (l,n-1)
    in Jvmtype.class {pkgs = pkgs, name = name'}
    end

fun Prl s = TextIO.print (s^"\n");


fun qualName (Jvmtype.CLASS c) = 
    let  (* From sml-jvm Jvmtype,  but modified for dots instead of slashes *)
	fun revAppend s cs = CharVector.foldl (op ::) cs s
	fun revAppendDelim (s, cs) = #"." :: (revAppend s cs)
	fun revQualName cs {pkgs, name} = revAppend name (foldl revAppendDelim cs pkgs)
    in
	String.implode(rev (revQualName [] c))
    end

fun tyToJty BOOLEANty    = Jvmtype.Tboolean
  | tyToJty CHARty       = Jvmtype.Tchar
  | tyToJty FLOATty      = Jvmtype.Tfloat
  | tyToJty DOUBLEty     = Jvmtype.Tdouble
  | tyToJty BYTEty       = Jvmtype.Tbyte
  | tyToJty SHORTty      = Jvmtype.Tshort
  | tyToJty INTty        = Jvmtype.Tint
  | tyToJty LONGty       = Jvmtype.Tlong
  | tyToJty (ARRAYty(t)) = Jvmtype.Tarray(tyToJty t)
  | tyToJty (REFty x) = Jvmtype.Tclass (qnameToJClass x)
  | tyToJty _ = raise gfError "UNKNOWN TYPE"

fun rtyToJty (SOME t) = SOME (tyToJty t)
  | rtyToJty NONE = NONE

fun jtyToTy Jvmtype.Tboolean    = BOOLEANty
  | jtyToTy Jvmtype.Tchar       = CHARty
  | jtyToTy Jvmtype.Tfloat      = FLOATty
  | jtyToTy Jvmtype.Tdouble     = DOUBLEty
  | jtyToTy Jvmtype.Tbyte       = BYTEty
  | jtyToTy Jvmtype.Tshort      = SHORTty
  | jtyToTy Jvmtype.Tint        = INTty
  | jtyToTy Jvmtype.Tlong       = LONGty
  | jtyToTy (Jvmtype.Tarray(t)) = ARRAYty(jtyToTy t) 
  | jtyToTy (Jvmtype.Tclass(x)) = REFty(qualName x)

fun jtyToRty (SOME t) = SOME (jtyToTy t)
  | jtyToRty NONE = NONE

fun  jTyDescToRTy s =
    if s = "void" then NONE
    else (SOME (jTyDescToTy s))

fun jTyDescToJty s = tyToJty (jTyDescToTy s)


fun mdescToMethodRef (MDESC(ty, name, argtypes)) = 
    let 
	val l = String.tokens (fn c => (c = #".")) name;
	val n = length l;
	val msig = (map tyToJty argtypes, rtyToJty ty);
	val name = nth(l,n-1);
	val (pkgs, cname) = 
	    if n=1 then ([],"")
	    else (take(l,n-2), nth(l,n-2));
    in 
	{name=name, msig=msig, class = Jvmtype.class{pkgs=pkgs, name=cname}}
    end

fun methodRefToMdesc (mref: Bytecode.method_ref) =
    let
	val {class=class, name=name, msig=msig} = mref
	val (args, rty) = msig
	val rty' = jtyToRty rty
	val name' = qualName(class) ^ "." ^ name
	val args' = map jtyToTy args
    in
	MDESC(rty', name', args')
    end

fun fdescToFieldRef (FDESC(fty,name)) =
  let val l = String.tokens (fn c => (c = #".")) name
      val n = length l
      val ty = tyToJty fty
      val name = nth(l,n-1)
      val (pkgs, cname) = 
	  if n=1 then ([],"")
	  else (take(l,n-2), nth(l,n-2))
  in 
      {name=name, ty=ty, class = Jvmtype.class{pkgs=pkgs, name=cname}}
  end;

fun fieldRefToFdesc (fref: Bytecode.field_ref) =
    let
	val {class=class, name=name, ty=ty} = fref
	val ty' = jtyToTy ty
	val name' = qualName class ^ "." ^ name
    in
	FDESC(ty', name')
    end

fun javatyToRTy (t: Type) =
    if t.#equals(Type.VOID) then NONE
	 else 	SOME (jTyDescToTy (valOf((t.#toString()))))
(********* DODGY *********)


fun javatyToTy (t: Type) = valOf(javatyToRTy t)

fun makeMethRef (cp: ConstantPoolGen) (j: InvokeInstruction) =
let
    val argTypes = arrayToList(valOf(j.#getArgumentTypes(cp)))
    val argTypes' = map (fn (x:Type) => javatyToTy(x)) argTypes
    val ret = javatyToRTy(valOf(j.#getReturnType(cp)))
    val cname = valOf(j.#getClassName(cp))
    val mname = valOf(j.#getMethodName(cp))
in
    mdescToMethodRef (MDESC(ret, cname^"."^mname, argTypes'))
end

fun makeFieldRef (cp: ConstantPoolGen) (j: FieldInstruction) =
let
    val ty = javatyToTy(valOf(j.#getFieldType(cp)))
    val cname = valOf(j.#getClassName(cp))
    val fname = valOf(j.#getFieldName(cp))
in
    fdescToFieldRef (FDESC(ty, cname^"."^fname))
end

fun jInstrHandleToAbsInstr (cp: ConstantPoolGen) (h: InstructionHandle) =
    let 
	val j = valOf(h.#getInstruction())
(*	val () = TextIO.print (valOf(h.#toString()))*)
(* Oddly,  if we operate directly on the list of instructions,  then 
   branch instructions don't seem to have any targets *)
(* OK, maybe it's not so odd *)

	fun ind (j: LocalVariableInstruction) =
	    Localvar.fromInt (j.#getIndex())

	local 
	    fun target(j: BranchInstruction) = 
		let 
		    val ih = case j.#getTarget() of 
			NONE => 
			    raise gfError ("Missing target in  " ^ valOf(j.#toString(true)))
		      | SOME ih => ih
		in findLabel ih 
		end
	in
	    fun getFunTarget (j: BranchInstruction) = 
		case target j of
		    FUNlabel l => l
		  | _ => raise gfError "Bad label for function call"

    	    fun getIfTarget (j: BranchInstruction) = 
		case target j of
		    IFlabel l => l
		  | _ => raise gfError "Bad label for non-function"
	end
    in

    if _instanceof ACONST_NULL j then Jaconst_null

    else if _instanceof ICONST j then
	let
	    val n: java.lang.Number = valOf ((j:> ICONST).#getValue())
	in
	    Jiconst(n.#intValue())
	end
    else if _instanceof BIPUSH j then
	let
	    val n: java.lang.Number = valOf ((j:> BIPUSH).#getValue())
	in
	    Jiconst(n.#intValue())
	end
    else if _instanceof SIPUSH j then
	let
	    val n: java.lang.Number = valOf ((j:> SIPUSH).#getValue())
	in
	    Jiconst(n.#intValue())
	end
    else if _instanceof LDC j then 
	let
	    val v : java.lang.Object = valOf ((j:> LDC).#getValue(cp))
	    val t : Type = valOf ((j:> LDC).#getType(cp) )
	in 
	    if _instanceof java.lang.Integer v then
	       Jiconst( (v:> java.lang.Integer).#intValue())
	    else if _instanceof java.lang.Float v then
		Jfconst((v:> java.lang.Float).#doubleValue())
	    else if _instanceof string v then
               Jsconst (v :> string)
	    else
		raise gfError "Unknown CP entry"
	end
    else if _instanceof LDC_W j then 
	let
	    val v : java.lang.Object = valOf ((j:> LDC_W).#getValue(cp))
	    val t : Type = valOf ((j:> LDC_W).#getType(cp) )
	in 
	    if _instanceof java.lang.Integer v then
	       Jiconst( (v:> java.lang.Integer).#intValue())
	    else if _instanceof java.lang.Float v then
	       Jfconst((v:> java.lang.Float).#doubleValue())
	    else if _instanceof string v then
               Jsconst (v :> string)
	    else
		raise gfError "Unknown CP entry"
	end
    else if _instanceof FCONST j then
	let
	    val n: java.lang.Number = valOf ((j:> FCONST).#getValue())
	in
	    Jfconst(n.#doubleValue())
	end
    else if _instanceof DUP j then Jdup
    else if _instanceof POP j then Jpop
    else if _instanceof ILOAD  j then Jiload (ind (j:> LocalVariableInstruction))
    else if _instanceof FLOAD j then Jfload (ind (j:> LocalVariableInstruction))
    else if _instanceof ALOAD  j then Jaload (ind (j:> LocalVariableInstruction))
    else if _instanceof ISTORE j then Jistore (ind (j:> LocalVariableInstruction))
    else if _instanceof FSTORE j then Jfstore (ind (j:> LocalVariableInstruction))
    else if _instanceof ASTORE j then Jastore (ind (j:> LocalVariableInstruction))
    else if _instanceof IADD j then Jiadd
    else if _instanceof ISUB j then Jisub
    else if _instanceof IMUL j then Jimul
    else if _instanceof IDIV j then Jidiv
    else if _instanceof IREM j then Jirem
    else if _instanceof FADD j then Jfadd
    else if _instanceof FSUB j then Jfsub
    else if _instanceof FMUL j then Jfmul
    else if _instanceof FDIV j then Jfdiv
    else if _instanceof FREM j then Jfrem
    else if _instanceof FCMPL j then Jfcmpl				    
    else if _instanceof IF_ACMPEQ j then Jif_acmpeq (getIfTarget (j:> BranchInstruction))
    else if _instanceof IF_ACMPNE j then Jif_acmpne (getIfTarget (j:> BranchInstruction))
    else if _instanceof IF_ICMPEQ j then Jif_icmpeq (getIfTarget (j:> BranchInstruction))
    else if _instanceof IF_ICMPNE j then Jif_icmpne (getIfTarget (j:> BranchInstruction))
    else if _instanceof IF_ICMPLT j then Jif_icmplt (getIfTarget (j:> BranchInstruction))
    else if _instanceof IF_ICMPLE j then Jif_icmple (getIfTarget (j:> BranchInstruction))
    else if _instanceof IF_ICMPGT j then Jif_icmpgt (getIfTarget (j:> BranchInstruction))
    else if _instanceof IF_ICMPGE j then Jif_icmpge (getIfTarget (j:> BranchInstruction))
    else if _instanceof IFEQ j then Jifeq (getIfTarget (j:> BranchInstruction))
    else if _instanceof IFNE j then Jifne (getIfTarget (j:> BranchInstruction))
    else if _instanceof IFLT j then Jiflt (getIfTarget (j:> BranchInstruction))
    else if _instanceof IFLE j then Jifle (getIfTarget (j:> BranchInstruction))
    else if _instanceof IFGT j then Jifgt (getIfTarget (j:> BranchInstruction))
    else if _instanceof IFGE j then Jifge (getIfTarget (j:> BranchInstruction))
    else if _instanceof GotoInstruction j then Jgoto (getFunTarget (j:> BranchInstruction))
    else if _instanceof ReturnInstruction j then Jreturn
    else if _instanceof INVOKEVIRTUAL j 
	     then Jinvokevirtual (makeMethRef cp (j:> InvokeInstruction))
    else if _instanceof INVOKESTATIC j 
	     then Jinvokestatic (makeMethRef cp (j:> InvokeInstruction))
    else if _instanceof INVOKESPECIAL j 
	     then Jinvokespecial (makeMethRef cp (j:> InvokeInstruction))
    else if _instanceof PUTFIELD j then Jputfield (makeFieldRef cp (j:> FieldInstruction)) 
    else if _instanceof GETFIELD j then Jgetfield (makeFieldRef cp (j:> FieldInstruction)) 
    else if _instanceof PUTSTATIC j then Jputstatic (makeFieldRef cp (j:> FieldInstruction)) 
    else if _instanceof GETSTATIC j then Jgetstatic (makeFieldRef cp (j:> FieldInstruction)) 
    else if _instanceof CHECKCAST j then
	let 
	    val lct: ObjectType = valOf((j:>CHECKCAST).#getLoadClassType(cp))
	    val cname = valOf(lct.#getClassName())
	in
	    Jcheckcast (Bytecode.CLASS(qnameToJClass cname))
	end

    else if _instanceof INSTANCEOF j then 
	let 
	    val lct: ObjectType = valOf((j:>INSTANCEOF).#getLoadClassType(cp))
	    val cname = valOf(lct.#getClassName())
	in
	    Jinstanceof (Bytecode.CLASS(qnameToJClass cname))
	end

    else if _instanceof NEW j then 
	let 
	    val lct: ObjectType = valOf((j:>NEW).#getLoadClassType(cp))
	    val cname = valOf(lct.#getClassName())
	in
	    Jnew (qnameToJClass cname)
	end
    else if _instanceof NEWARRAY j then
	let
	    val ty = valOf((j:> NEWARRAY).#getType())
	    val str = valOf(ty.#toString())
	    val jty = jTyDescToJty str
	in
	    Jnewarray { elem = jty, dim = 1 } 
	end
    else if _instanceof ANEWARRAY j then 
	let
	    val ty = valOf((j:> ANEWARRAY).#getType(cp))
	    val str = valOf(ty.#toString())
	    val jty = jTyDescToJty str
	in
	    Jnewarray { elem = jty, dim = 1 } 
	end
    else if _instanceof MULTIANEWARRAY j then 
	let
	    val jj = j:> MULTIANEWARRAY
	    val ty = valOf (jj.#getType(cp))
	    val str = valOf (ty.#toString())
	    val jty = jTyDescToJty str
	in
	    Jnewarray { elem = jty, dim = Int16.toInt (jj.#getDimensions()) }
	end
    else if _instanceof IALOAD j then Jiaload
    else if _instanceof FALOAD j then Jfaload
    else if _instanceof AALOAD j then Jaaload
    else if _instanceof IASTORE j then Jiastore
    else if _instanceof FASTORE j then Jfastore
    else if _instanceof AASTORE j then Jaastore
    else if _instanceof ARRAYLENGTH j then Jarraylength
    else if _instanceof F2I j then Jf2i
    else if _instanceof I2F j then Ji2f

(* Following section not needed for Grail code. Experimental. *)

    else if _instanceof IINC j then 
	let
	    val inc = (j :>IINC).#getIncrement()
	    val v = ind (j :>LocalVariableInstruction)
	in
	    Jiinc {var = v, const = inc}
	end
    else if _instanceof IFGE j then Jifge (getIfTarget (j:> BranchInstruction))

(* /Experimental *)


    else (TextIO.print ("* " ^ (valOf((j.#toString()))) ^ "\n"); Jnop)
end

fun hd [] = raise gfError "Empty list"
  | hd (h::t) = h


fun findvar n =
    let fun find n [] = NONE
	  | find n ((m,nm,ty)::t) = 
	    if m=n 
		then SOME (m,nm,ty) 
	    else find n t
    in
	find n (!varInfo)
    end

fun nameOfVar n = 
    let 
	val n' = Localvar.toInt(n)
    in 
	case findvar n' of 
	    NONE => "Var" ^ Int.toString(n')  (* ERROR *)
	  | SOME (_,nm,_) => nm
    end

fun typeAndNameOfVar n = 
	case findvar n of 
	    NONE => (INTty, "COMPILER BUG")
		(* raise gfError ("Var" ^ Int.toString(n) ^ " not found") *)
	  | SOME (_,nm,ty) => (ty,nm)
	    
(* fun nameOfFun (Label.LBL n) = "F" ^ Int.toString(n) *)

fun typeOfVar var = 
    let fun find n [] = NONE
	  | find n ((m,nm,ty)::t) = 
	    if nm=n 
		then SOME (m,nm,ty) 
	    else find n t
    in
	case find var (!varInfo) of
	    NONE => REFty "COMPILER ERROR"
	  | SOME (_,_,ty) => ty
    end

fun nameOfFun l funInfo = 
let 
    fun pos _ [] _ = raise gfError "Missing function label"
      | pos x (y::t) n = if x=y then n else pos x t n+1 

    val n = pos l (getFunLabels()) 0
    val (name, params) = nth (funInfo, n)
in 
    (name, map snd params)
end

fun chkStk s n = 
let
    val len = length s 
in
    if len < n then raise gfError "Too few items on stack"
    else if len = n then ()
	 else raise gfError "Too many items on stack"
end


(* Now we're trying to parse a fun body;  this involves collecting
   let decs then the result of the function.  This is tricky because 
   it's difficult to tell where the decs end and the result begins.
   The functions below scan the bytecode looking for primops then
   do some lookahead to try and work out what's coming.  The "Thing" 
   datatype below describes what's happening.  It's particularly
   messy when the result of the function is "if ..." because by the
   time we see the if we've already got two values on the stack,
   which might also have been args for a primop.  *)


datatype Thing =
    VAL of PrimOp 
  | VOID of PrimOp
  | NOTHING
  | IF of Value * Value * Test * Label.label

fun chop [] = raise gfError "Premature end of bytecode"
  | chop (h::t) = (h,t)

fun getPrimOp C =
    let 
	(* val () = prInstr (hd C) *)
	fun mkBinOp oper stk = (* Poor error checking *)
	    case stk of 
		[] => raise gfError "Too few arguments for BinOp"
	      | [_] => raise gfError "Too few arguments for BinOp"
	      | [x, y] =>BINop(oper, x, y) 
	      | _ => raise gfError "Too many arguments for BinOp"

	fun getVar [] = raise gfError "Looking for value on empty stack"
	  | getVar (h::_) = 
	    let in case h of
		VARval v => v
	      | _ => raise gfError "Seeking variable on stack,  but found something else"
	    end
    
	fun mkIF stk tst label = 
	    case stk of
		[]  => raise gfError "Too few arguments for test"
	      | [_] => raise gfError "Too few arguments for test"
	      | [NULLval _, NULLval _] => IF(NULLval "java.lang.Object",
					     NULLval "java.lang.Object",
					     tst, label)
	      | [NULLval _, VARval y] => IF(NULLval (tyToString (typeOfVar y)), 
					    VARval y, tst, label)	       
	      | [VARval x, NULLval _] => IF(VARval x, 
					    NULLval (tyToString(typeOfVar x)), tst, label)
	      | [x,y] => IF(x, y, tst, label)
	      | _ => raise gfError "Too many arguments for test"

	fun mkGET stk =
	    case stk of 
		[]  => raise gfError "Too few arguments for array load"
	      | [_] => raise gfError "Too few arguments for array load"
	      | [arr,i] => VAL(GETop(arr,i))
	      | _ => raise gfError "Too many arguments for array load"

	fun mkSET stk =
	    case stk of 
		[VARval(arr),i,NULLval _] => 
		let val x = 
		    (case typeOfVar arr of
			 ARRAYty ty => tyToString ty
		       | _ => "java.lang.Object")
		in 
		    VOID(SETop(VARval(arr),i,NULLval x))
		end
	      | [arr,i,a] => VOID(SETop(arr,i,a))
	      | _ => raise gfError "Wrong number of arguments for array store"

	fun eatValues C stk =
	    let
		val (h,t) = chop C
(*		val _ = (TextIO.print ("eat "); prInstr h) *)
	    in
		case h of 
		    Jiload n  => eatValues t (stk@[VARval (nameOfVar n)])
		  | Jfload n  => eatValues t (stk@[VARval (nameOfVar n)])
		  | Jaload n  => eatValues t (stk@[VARval (nameOfVar n)])
		  | Jiconst n => eatValues t (stk@[INTval n])
		  | Jfconst n => eatValues t (stk@[FLOATval n])
		  | Jsconst n => eatValues t (stk@[STRINGval n])
		  | Jaconst_null => eatValues t (stk@[NULLval ""])  (*** DANGER ****)
		  | Jdup => eatValues t (stk)
		  | _ => (C, stk)
	    end

	val (D, stk) = eatValues C []
	val (h,C') = chop D
(*	val _ = (TextIO.print (Int.toString(length stk)^ " > "); prInstr h) *)
    in
	case h of
	    Jiadd => (VAL(mkBinOp ADDop stk), C')
	  | Jisub => (VAL(mkBinOp SUBop stk), C')
	  | Jimul => (VAL(mkBinOp MULop stk), C')
	  | Jidiv => (VAL(mkBinOp DIVop stk), C')
	  | Jirem => (VAL(mkBinOp MODop stk), C')

	  | Jfadd => (VAL(mkBinOp ADDop stk), C')
	  | Jfsub => (VAL(mkBinOp SUBop stk), C')
	  | Jfmul => (VAL(mkBinOp MULop stk), C')
	  | Jfdiv => (VAL(mkBinOp DIVop stk), C')
	  | Jfrem => (VAL(mkBinOp MODop stk), C')

	  | Ji2f => 
	    let 
		val () = chkStk stk 1
		val v = VARval(getVar stk)
	    in 
		(VAL(ITOFop(v)), C')
	    end

	  | Jf2i => 
	    let 
		val () = chkStk stk 1
		val v = VARval(getVar stk)
	    in 
		(VAL(FTOIop(v)), C')
	    end

	  | Jarraylength => 		
	    (case stk of
		 [obj] => (VAL(LENGTHop(obj)), C')
	       | _ => raise gfError "Must be 1 value on stack"
	    )

	  | Jnew class => 
	    let 
		val () = chkStk stk 0
		val () = let 
			     val d = hd C' 
			 in case d of
			     Jdup => ()
			   | _ => raise gfError "No dup after new"
			 end
		(* consume the constructor args *)
		val pop = getPrimOp ((Jiload(Localvar.fromInt 0)) :: (tl C'))
		    (* This is a bit nasty;  invokespecial requires 
		       an object on the stack.  We've only got a dup,
		       so let's replace it with an arbitrary iload 0
		       (which is going to be discarded anyway) *)
	    in
		let in case pop of 
		   (VOID(INVOKESPECIALop(v,md,args)), C'') => (VAL(NEWop(md,args)), C'')
		 | (VAL _, _) => raise gfError "Constructor returns value"
		 | _ => raise gfError "Expecting invokespecial after new"
		end
	    end

	  | Jcheckcast cref => 
		let 
		    val () = chkStk stk 1 
		    val class = case cref of 
			CLASS jclass => jclass
		      | ARRAY _ => raise gfError "Trying to cast to array type"
		in
		    (VAL(CHECKCASTop(qualName class, getVar stk)), C')
		end 
	    
	  | Jinstanceof cref => 
		let 
		    val () = chkStk stk 1 
		    val class = case cref of 
			CLASS jclass => jclass
		      | ARRAY _ => raise gfError "Trying to call 'instanceof' on array type"
		in
		    (VAL(INSTANCEop(qualName class, getVar stk)), C')
		end 
	    
	  | Jinvokestatic mref =>
		let 
		    val mdesc = methodRefToMdesc mref
		    val (MDESC(rty,_,argtys)) = mdesc
		    val () = chkStk stk (length argtys)
		in
		    case rty of 
			NONE => (VOID(INVOKESTATICop (mdesc, stk)), C')
		      | _    => (VAL(INVOKESTATICop (mdesc, stk)), C')
		end 
       
	  | Jinvokevirtual mref =>
		let 
		    val mdesc = methodRefToMdesc mref
		    val (MDESC(rty,_,argtys)) = mdesc
		    val () = chkStk stk (1+length argtys)
		    val obj = getVar stk
		    val args = tl stk
		in
		    case rty of 
			NONE => (VOID(INVOKEVIRTUALop (obj, mdesc, args)), C')
		      | _    => (VAL (INVOKEVIRTUALop (obj, mdesc, args)), C')
		end
	    
	  | Jinvokespecial mref =>
		let 
		    val mdesc = methodRefToMdesc mref
		    val (MDESC(rty,_,argtys)) = mdesc
		    val () = chkStk stk (1+length argtys)
		    val obj = getVar stk
		    val args = tl stk
		in
		    case rty of 
			NONE => (VOID(INVOKESPECIALop (obj, mdesc, args)), C')
		      | _    => (VAL (INVOKESPECIALop (obj, mdesc, args)), C')
		end
	    
	  | Jdup => raise gfError "Unexpected dup opcode"
	  | Jgetstatic fref =>
		let 
		    val () = chkStk stk 0
		in
		    (VAL(GETSTATICop (fieldRefToFdesc fref)), C')
		end
	    
	  | Jgetfield fref => 
		let 
		    val () = chkStk stk 1
		    val obj = getVar stk
		in
		    (VAL(GETFIELDop (obj, fieldRefToFdesc fref)), C')
		end
	    
	  | Jputstatic fref =>
		let 
		    val () = chkStk stk 1
		    val input = hd stk
		in
		    (VOID(PUTSTATICop (fieldRefToFdesc fref, input)), C')
		end
	    
	  | Jputfield fref =>
		let 
		    val () = chkStk stk 2
		    val obj = getVar stk
		    val input = hd (tl stk)
		in
		    (VOID(PUTFIELDop (obj, fieldRefToFdesc fref, input)), C')
		end

	  | Jlabel _ => getPrimOp C'  (* Slight problem because first and 
				         last instructions have labels *)

	  | Jif_acmpeq l => (mkIF stk EQtest l, C')
	  | Jif_acmpne l => (mkIF stk NEtest l, C')

	  | Jif_icmpeq l => (mkIF stk EQtest l, C')
	  | Jif_icmpne l => (mkIF stk NEtest l, C')
	  | Jif_icmplt l => (mkIF stk Ltest l, C')
	  | Jif_icmple l => (mkIF stk LEtest l, C')
	  | Jif_icmpgt l => (mkIF stk Gtest l, C')
	  | Jif_icmpge l => (mkIF stk GEtest l, C')

	  | Jfcmpl => 
	    let 
		val nxt = hd C'
		val C'' = tl C'
	    in case nxt of 
		Jifeq l => (mkIF stk EQtest l, C'')
	      | Jifne l => (mkIF stk NEtest l, C'')
	      | Jiflt l => (mkIF stk Ltest l, C'')
	      | Jifle l => (mkIF stk LEtest l, C'')
	      | Jifgt l => (mkIF stk Gtest l, C'')
	      | Jifge l => (mkIF stk GEtest l, C'')
	      | _ => raise gfError "fcmpl must be followed by if---"
	    end
	  | Jifeq l => raise gfError "Found ifeq  without fcmpl"
	  | Jifne l => raise gfError "Found ifne  without fcmpl"
	  | Jiflt l => raise gfError "Found iflt  without fcmpl"
	  | Jifle l => raise gfError "Found ifle  without fcmpl"
	  | Jifgt l => raise gfError "Found ifgt  without fcmpl"
	  | Jifge l => raise gfError "Found ifge  without fcmpl"
	  | Jiaload => (mkGET stk, C')
	  | Jfaload => (mkGET stk, C')
	  | Jaaload => (mkGET stk, C')
	  | Jiastore => (mkSET stk, C')
	  | Jfastore => (mkSET stk, C')
	  | Jaastore => (mkSET stk, C')
	  | Jnewarray {elem,dim} => 
	    let 
		val () = chkStk stk 1
	    in
	    case dim of 
		1 => ((VAL (EMPTYop(hd stk, jtyToTy elem))), C')
	      | _ => raise gfError "Only one-dimensional arrays yet. Give me a minute."
	    end

(* Experimental  *)

          | Jiinc {var, const} => 
	    let 
		val () = chkStk stk 0
	    in
		(VAL (BINop(ADDop, VARval(nameOfVar var), INTval(const))), C')
	    end

(* /Experimental *)


	  | _ => let in case stk of
		[] => (NOTHING, D)
	      | [x] => (VAL (VALop x), D)
	      | _ => raise gfError "Too many values on stack"
	    end
    end

fun getFunDec C decs funInfo = 
    let 
	fun getRes1 Jreturn = VOIDres
	  | getRes1 (Jgoto n) = FUNres (nameOfFun n funInfo)
	  | getRes1 x = (prInstr x; raise gfError "Malformed result")

	fun getRes q i =
	    case q of 
		VOID p => 
		    let in case i of 
			Jreturn => OPres p
		      | _ => raise gfError "Missing return"
		    end
	      | VAL p =>  
		    let in case i of 
			Jreturn => OPres p
		      | _ => raise gfError "Missing return"
		    end
	      | NOTHING => getRes1 i
	      | _ => raise gfError "Malformed result (nested if?)"

	val (q,D) = getPrimOp C
(*	val _ = TextIO.print "Got primop\n" *)
	val (h,D') = chop D
    in
	case q of
	    VOID p => getFunDec D (decs@[VOIDdec p]) funInfo

	  | VAL p => 
		let in case h of
		    Jistore n => getFunDec D' (decs@[VALdec(nameOfVar n, p)]) funInfo
		  | Jastore n => 
		    let val x = (case p of 
				 VALop(NULLval _) => 
				 VALdec(nameOfVar n, 
					VALop(NULLval (tyToString (typeOfVar (nameOfVar n)))))

			       | _ => VALdec(nameOfVar n, p))
		    in 
			getFunDec D' (decs@[x]) funInfo
		    end
		  | Jastore n => getFunDec D' (decs@[VALdec(nameOfVar n, p)]) funInfo
		  | Jfstore n => getFunDec D' (decs@[VALdec(nameOfVar n, p)]) funInfo
		  | Jreturn => (decs, PRIMres(OPres p), D')
		  | _ => raise gfError "Unused value on stack"
		end

	  | NOTHING => (decs, PRIMres(getRes1 h), D')

	  | IF  (x, y, tst, dest) =>
		let
		    val (r,E) = getPrimOp D
		    val (h,t) = chop E
		    val p1 = getRes r h
		    val (h',t') = chop t
		    val () = let in case h' of 
			Jlabel l => if l<>dest 
					then raise gfError "Bad destination in 'if ...'"
				    else ()
		      | _ => raise gfError "Missing label in 'if ...'"
			     end
		    val (r',E') = getPrimOp t'
		    val (h'',t'') = chop E'  
		    val p2 = getRes r' h''
		in
		    (decs, CHOICEres(x,tst,y,p1,p2), t'')  
			(* This was getting things the wrong way round *)
		end
    end    


fun getFunDecs C funInfo =
    let
	fun get [] acc _ = acc
	  | get C acc first =
	    let
		val (l,C') = chop C
		val _ = case l of 
		    Jlabel n => ()  (* DON'T REALLY NEED THIS [MAYBE] *)
		  | _ => if first then () (* first funblock has no name *) (* YUCK *)
			 else raise gfError "Missing label at start of function"
		val (decs, result, rest) = getFunDec C [] funInfo
(*val _ = TextIO.print ("got fundec\n")*)
	    in 
		get rest (acc@[(decs,result)]) false
	    end
    in
	get C [] true
    end


fun getFunInfo attrs =   (* ONLY call this AFTER we've set varInfo *)
                         (* because typeAndNameOfVar depends on contents of varInfo *)
let

    (* Look for MRG function info attribute *)

    fun findMRGfunInfo [] = Array.fromList []
      | findMRGfunInfo ((h: Attribute)::t) = 
	if _instanceof Unknown h
	    then
		let 
		    val u = h:>Unknown
		    val name = valOf(u.#getName())
		in
		    case name of 
			"uk.ac.ed.dcs.mrg.GrailFunctionInfo" => 
			    let in case u.#getBytes() of 
				NONE => Array.fromList []
			      | SOME b => b
			    end
		      | _ => findMRGfunInfo t
		end
	else findMRGfunInfo t
	    
    fun array8ToList a =
	let fun g a j l =
	    if j < 0 then l
	    else g a (j-1) ((Int8.toInt(Array.sub(a,j)))::l)
	in
	    g a (Array.length(a)-1) []
	end

    (* Convert list of bytes into (fun_name * (type*param_name)list)list *)
    fun deserialise [] = []
      | deserialise l =
	let 
	    fun getStr [] = ("", [])
	      | getStr (h::t) =
		let 
		    val x = take(t,h)
		in 
		    (String.implode (map Char.chr x), drop(t,h))
		end
	    
	    fun getArgs [] = ([],[])
	      | getArgs (h::t) =
		    (map typeAndNameOfVar (take(t,h)), drop(t,h))


	    val (id, t) = getStr l
	    val (args, t') = getArgs t
(*	    val _ = TextIO.print ((Int.toString h) ^ " -> " ^ id ^"\n")*)
	in
	    (id, args)::(deserialise t')
	end


    fun p (l,i,m) = 
	let val pr = TextIO.print
	in
	    pr (Int.toString l);
	    pr ( " -> ");
	    pr i;
	    pr ": ";
	    app (fn x => pr (Int.toString(x) ^ " ")) m;
	    pr "\n"
	end
in
    deserialise (array8ToList (findMRGfunInfo attrs))
end

	

fun jmethodToMethodDef className (cp: ConstantPoolGen) (m: Method) =
let
    fun translateHandles [] = []
      | translateHandles (h::t) = case lookupLabel h of 
             NONE => (jInstrHandleToAbsInstr cp h)::(translateHandles t)
           | SOME (FUNlabel p) 
	     => (Jlabel p) :: (jInstrHandleToAbsInstr cp h)::(translateHandles t)
           | SOME (IFlabel p) 
	     => (Jlabel p) :: (jInstrHandleToAbsInstr cp h)::(translateHandles t)

    val () = resetNames()

    val sign = valOf (m.#getSignature())
    val ret = valOf(Utility.methodSignatureReturnType(sign, false))
    val jargs = arrayToList(valOf(Utility.methodSignatureArgumentTypes(sign, false)))
	(* false -> don't chop argument types *)
    val argtypes = map (fn x => jTyDescToTy x) jargs
    val flags = getFlags (m:>AccessFlags)


    val mg = _new MethodGen(m, SOME className, cp)
	
    val attrs = arrayToList(valOf(mg.#getCodeAttributes()))

    val ihs = case mg.#getInstructionList() of 
	NONE => []  (* native method for example *)
	  | SOME il => arrayToList (valOf(il.#getInstructionHandles()))

    val lvg = mg.#getLocalVariables()
    val lvars = arrayToList(valOf lvg)

    fun pr x = TextIO.print x
    fun prl x = TextIO.print (x ^ "\n")
	    
    fun getVarInfo (x: LocalVariableGen) = 
	let 
	    val t = valOf(x.#getType()) 
	in
	    (x.#getIndex(), valOf(x.#getName()), jTyDescToTy (valOf((t.#toString()))))
 	end

    val _ = (varInfo := (map getVarInfo lvars))

    val argnames = 
	let
	    val offset = if m.#isStatic() then 0 else 1
	    val n = length argtypes
	in
	    tabulate (n, (fn i => (nameOfVar(Localvar.fromInt(i+offset)))))
	end

    val funInfo = getFunInfo attrs


    fun prVarDesc (x: LocalVariableGen) = 
	let 
	    val t = valOf(x.#getType()) 
	in
	    prl("Variable " 
		^ (Int.toString(x.#getIndex()))
		^ " is " 
		^ (valOf(x.#getName()))
		^ " : " 
		^ (valOf(t.#toString())))
	end

    val () = resetLabels()
    val () = getTargets ihs
    val () = setFunLabels()
	
    val code    = translateHandles ihs
    val funDecs = getFunDecs code funInfo

    val ((decs1, result1), rest) = 
	case funDecs of [] => raise gfError "Oops - no funDecs at all"
      | _ => (hd funDecs, tl funDecs)

    val funs = 
	if (length rest) <> (length funInfo) then raise gfError "Corrupt metadata"
	else (* WOW - this is really bad *)
	    let 
		val bodies = map (fn (d,r) => FUNbody(d,r)) rest
		fun makeFdec ((id,args), body) = FDEC(id, args, body)
	    in 
		ListPair.map makeFdec (funInfo, bodies)
	    end

in
    MDEF(flags, 
	 jTyDescToRTy ret, 
	 valOf(m.#getName()), 
	 ListPair.zip(argtypes, argnames),
	 MBODY(decs1, funs, result1)) 
         (* tidy this up *)
end


fun decompile (fname: string) = 
let
    val class = 
	case (org.apache.bcel.Repository.lookupClass fname) of 
	    NONE => raise gfError ("Can't find class " ^ fname)
	  | SOME c => c
		    
    val className = valOf (class.#getClassName())
    val cp = _new ConstantPoolGen(class.#getConstantPool())

    val attrs = valOf (class.#getAttributes())
    val fields = arrayToList (valOf (class.#getFields()))
    val fds = map jfieldToFieldDef fields
    val methods = arrayToList(valOf (class.#getMethods()))
(*
    val methods = filter (fn (m: Method) => (valOf(m.#getName()) <> "<init>")) methods 
    (* Throw away initialisers for now *)
*)
    val mds = map (jmethodToMethodDef className cp) methods
    val cd = CDEF([], className, fds, mds)
in
    prClassDef cd
end



fun main (env : string option array option) =
  case env of 
    NONE => 
    TextIO.print "Please specify a Java class name\n"

  | SOME env' => 
    if Array.length env' = 0 then   TextIO.print "Please specify a Java class name\n"
    else
      case Array.sub(env', 0) of
        NONE => 
	    TextIO.print "Please specify a Java class name\n"
      | SOME fname => decompile fname

end (* of struct *)
