(* Classfile.sml
 *
 * Peter Bertelsen
 * December 1997
 *)

open Classdecl Constpool

(* NOTE: it is assumed that Constpool.indexValue and the labelMap
 * returned by Emitcode.emit return integers in the range
 * [0, 0xffff].
 *)

exception InvalidClass of string

fun classError s = raise InvalidClass("Classfile." ^ s)

fun bug s = raise Bytecode.InternalError("Classfile." ^ s)

type exn_hdl =
    {start : Word16.word,
     stop  : Word16.word,
     entry : Word16.word,
     catch : index option}

type line_number_info =
    {start : Word16.word,
     line  : Word16.word}

type local_var_info =
    {start  : Word16.word,
     length : Word16.word,
     name   : index,
     desc   : index,
     index  : Word16.word}

datatype attribute =
    SRCFILE of {attr : index,
		file : index}
  | CONSTVAL of {attr  : index,
		 value : index}
  | CODE of {attr   : index,
	     stack  : Word16.word,
	     locals : Word16.word,
	     code   : Word8Vector.vector,
	     hdls   : exn_hdl list,
	     attrs  : attribute list}
  | EXNS of {attr : index,
	     exns : index list}
  | LINENUM of {attr  : index,
		lines : line_number_info list}
  | LOCALVAR of {attr : index,
		 vars : local_var_info list}
  | ATTR of {attr : index,
	     info : Word8Vector.vector}
	
type member =
    {flags : Word16.word,
     name  : index,
     desc  : index,
     attrs : attribute list}

type class_file =
    {magic   : Word32.word,
     minor   : Word16.word,
     major   : Word16.word,
     pool    : pool,
     flags   : Word16.word,
     this    : index,
     super   : index option,
     ifcs    : index list,
     fields  : member list,
     methods : member list,
     attrs   : attribute list}

val magic' = Word8Vector.fromList[0wxca, 0wxfe, 0wxba, 0wxbe]
val magic  = valOf(Word32.fromBytes magic')
val minor  = Word16.fromWord 0w3
val major  = Word16.fromWord 0w45

fun toWord' flag =
	(case flag of
	     ACCpublic       => 0wx0001
	   | ACCprivate      => 0wx0002
	   | ACCprotected    => 0wx0004
	   | ACCstatic       => 0wx0008
	   | ACCfinal        => 0wx0010
	   | ACCvolatile     => 0wx0040
	   | ACCtransient    => 0wx0080
	   | ACCsynchronized => 0wx0020
	   | ACCnative       => 0wx0100
	   | ACCabstract     => 0wx0400
	   | ACCsuper        => 0wx0020   (* same as ACCsynchronized *)
	   | ACCinterface    => 0wx0200)

fun toWord (flag, w) = Word.orb(toWord' flag, w)

val flagsToWord = Word16.fromWord o (List.foldl toWord 0w0)

val isU2 = Emitcode.isU2

fun fromClassDecl cp {flags, this, super, ifcs, fdecls, mdecls, attrs} =
    let val insUtf8'  = insUtf8  cp
	val insConst' = insConst cp
	val insClass' = insClass cp

	fun insExnHandler labelMap {start, stop, entry, catch} =
	    let val labelMap' = Word16.fromInt o labelMap
	    in
		{start = labelMap' start,
		 stop  = labelMap' stop,
		 entry = labelMap' entry,
		 catch = (case catch of
			      NONE       => NONE
			    | SOME class => SOME (insClass' class))}
	    end

	fun insLnumInfo labelMap {start, line} =
	    if isU2 line then
		{start = Word16.fromInt(labelMap start),
		 line  = Word16.fromInt line}
	    else classError
		 "fromClassDecl.insLnumInfo: line number out of range"
	    
	fun insLvarInfo labelMap {from, thru, name, ty, index} =
	    let val start' = labelMap from
		val length = labelMap thru - start'
		val index' = Localvar.toInt index
	    in
		if isU2 length then
		    if isU2 index' then
			{start  = Word16.fromInt start',
			 length = Word16.fromInt length,
			 name   = insUtf8' name,
			 desc   = insUtf8'(Jvmtype.typeDesc ty),
			 index  = Word16.fromInt index'}
		    else classError
			"fromClassDecl.insLvarInfo: index out of range"
		else classError
		     "fromClassDecl.insLvarInfo: length out of range" 
	    end

	fun insAttr' _ _ (Classdecl.SRCFILE file) =
	        SRCFILE {attr = insUtf8' "SourceFile",
			 file = insUtf8' file}
	  | insAttr' _ _ (Classdecl.CONSTVAL value) =
		CONSTVAL {attr  = insUtf8' "ConstantValue",
			  value = insConst' value}
	  | insAttr' _ ty (Classdecl.CODE {stack, locals, code, hdls, attrs}) =
		let val (code', labelMap) = Emitcode.emit cp ty code
		    val insExnHandler'    = insExnHandler labelMap
		    val insAttr''         = insAttr' labelMap ty
		in
		    if isU2 stack then
			if isU2 locals then
			    CODE {attr   = insUtf8' "Code",
				  stack  = Word16.fromInt stack,
				  locals = Word16.fromInt locals,
				  code   = code',
				  hdls   = List.map insExnHandler' hdls,
				  attrs  = List.map insAttr'' attrs}
			else classError 
			     "insAttr [CODE]: max_locals out of range"
		    else classError "insAttr [CODE]: max_stack out of range"
		end
	  | insAttr' _ _ (Classdecl.EXNS exns) =
		EXNS {attr = insUtf8' "Exceptions",
		      exns = List.map insClass' exns}
	  | insAttr' labelMap _ (Classdecl.LINENUM lines) =
		LINENUM {attr  = insUtf8' "LineNumberTable",
			 lines = List.map (insLnumInfo labelMap) lines}
	  | insAttr' labelMap _ (Classdecl.LOCALVAR vs) =
		LOCALVAR {attr = insUtf8' "LocalVariableTable",
			  vars = List.map (insLvarInfo labelMap) vs}
	  | insAttr' _ _ (Classdecl.ATTR {attr, info}) =
		ATTR {attr = insUtf8' attr,
		      info = info}

	fun emptyLblMap _ = classError
	                    "fromClassDecl.insAttr: invalid nested attributes"

	val insAttr = insAttr' emptyLblMap

	    (* NOTE: when insAttr is used for inserting a method
	     * attribute, the second argument specifies the (optional)
	     * return type for the method; when insAttr is used for
	     * inserting a field or class attribute, the faked `return
	     * type' is NONE.
	     *)

	fun insField {flags, name, ty, attrs} =
	    {flags = flagsToWord flags,
	     name  = insUtf8' name,
	     desc  = insUtf8'(Jvmtype.typeDesc ty),
	     attrs = List.map (insAttr NONE) attrs}

	fun insMethod {flags, name, msig, attrs} =
	    let val (_, returnTy) = msig
	    in
		{flags = flagsToWord flags,
		 name  = insUtf8' name,
		 desc  = insUtf8'(Jvmtype.methodDesc msig),
		 attrs = List.map (insAttr returnTy) attrs}
	    end
    in
	(* NOTE: cp need not be completed before we start building the
	 * resulting class_file since Constpool.pool is imperative. *)
	{magic   = magic,
	 minor   = minor,
	 major   = major,
	 pool    = cp,
	 flags   = flagsToWord flags,
	 this    = insClass' this,
	 super   = (case super of
			NONE => NONE
		      | SOME class => SOME (insClass' class)),
	 ifcs    = List.map insClass' ifcs,
	 fields  = List.map insField fdecls,
	 methods = List.map insMethod mdecls,
	 attrs   = List.map (insAttr NONE) attrs}
    end

fun toClassDecl cf = raise Bytecode.Unimplemented "Classfile.toClassDecl"

val word16_0 = Word16.fromWord 0w0
val word32_2 = Word32.fromWord 0w2

fun emitClassFile out {magic, minor, major, pool, flags, this,
		       super, ifcs, fields, methods, attrs} =
    let val emitU2      = Word16.emit out
	val emitU2i     = emitU2 o Word16.fromInt
	val emitU4      = Word32.emit out
	val emitU4i     = emitU4 o Word32.fromInt
	val emitCpIndex = emitU2i o Constpool.indexValue

	fun emitCpIndexOpt  NONE    = emitU2 word16_0
	  | emitCpIndexOpt (SOME i) = emitCpIndex i

	fun emitExnHdl {start, stop, entry, catch} =
	    (emitU2 start;
	     emitU2 stop;
	     emitU2 entry;
	     emitCpIndexOpt catch)

	fun emitLineNumInfo {start, line} =
	        (emitU2 start;
		 emitU2 line)

	fun emitLocalVarInfo {start, length, name, desc, index} =
	        (emitU2 start;
		 emitU2 length;
		 emitCpIndex name;
		 emitCpIndex desc;
		 emitU2 index)

	fun attrSize (LINENUM {lines, ...}) = 8 + List.length lines * 4
	  | attrSize (LOCALVAR {vars, ...}) = 8 + List.length vars * 10
	  | attrSize (ATTR {info, ...}) = 2 + Word8Vector.length info
	  | attrSize _ = classError "emit.attrSize: invalid nested attributes"

	fun attrsSize attrs =
	    List.foldl (fn (a, sz) => attrSize a + sz) 0 attrs

	fun emitAttr (SRCFILE {attr, file}) =
	        (emitCpIndex attr;
		 emitU4 word32_2;
		 emitCpIndex file)
	  | emitAttr (CONSTVAL {attr, value}) =
		(emitCpIndex attr;
		 emitU4 word32_2;
		 emitCpIndex value)
	  | emitAttr (CODE {attr, stack, locals, code, hdls, attrs}) =
	        let val codeLength = Word8Vector.length code
		    val hdlsCount  = List.length hdls
		in
		    if isU2 hdlsCount then
			(emitCpIndex attr;
			 emitU4i(2 + 2 + 4 + codeLength +
				 2 + hdlsCount * 8 +
				 2 + attrsSize attrs);
			 emitU2 stack;
			 emitU2 locals;
			 emitU4i codeLength;
			 Word8Vector.app out code;
			 emitU2i hdlsCount;
			 List.app emitExnHdl hdls;
			 emitAttrs attrs)
		    else classError
			 "emit.emitAttr [CODE]: too many exception handlers"
		end
	  | emitAttr (EXNS {attr, exns}) =
		let val exnCount = List.length exns
		in
		    if isU2 exnCount then
			(emitCpIndex attr;
			 emitU4i(2 + exnCount * 2);
			 emitU2i exnCount;
			 List.app emitCpIndex exns)
		    else classError "emit.emitAttr [EXNS]: too many exceptions"
		end
	  | emitAttr (LINENUM {attr, lines}) =
		let val lineCount = List.length lines
		in
		    if isU2 lineCount then
			(emitCpIndex attr;
			 emitU4i(2 + lineCount * 4);
			 emitU2i lineCount;
			 List.app emitLineNumInfo lines)
		    else classError "emit.emitAttr [LINENUM]: too many lines"
		end
	  | emitAttr (LOCALVAR {attr, vars}) =
		let val varCount = List.length vars
		in
		    if isU2 varCount then
			(emitCpIndex attr;
			 emitU4i(2 + varCount * 10);
			 emitU2i varCount;
			 List.app emitLocalVarInfo vars)
		    else classError
			 "emit.emitAttr [LOCALVAR]: too many variables"
		end
	  | emitAttr (ATTR {attr, info}) =
		(emitCpIndex attr;
		 Word8Vector.app out info)
	and emitAttrs attrs =
	    let val len = List.length attrs
	    in
		if isU2 len then (emitU2i len;
				  List.app emitAttr attrs)
		else classError "emit.emitAttrs: too many attributes"
	    end

	fun emitMember {flags, name, desc, attrs} =
	    (emitU2 flags;
	     emitCpIndex name;
	     emitCpIndex desc;
	     emitAttrs attrs)

	val ifcsCount    = List.length ifcs
	val fieldsCount  = List.length fields
	val methodsCount = List.length methods
    in
	if isU2 ifcsCount then
	    if isU2 fieldsCount then
		if isU2 methodsCount then
		    (emitU4 magic;
		     emitU2 minor;
		     emitU2 major;
		     Constpool.emit out pool;
		     emitU2 flags;
		     emitCpIndex this;
		     emitCpIndexOpt super;
		     emitU2i ifcsCount;
		     List.app emitCpIndex ifcs;
		     emitU2i fieldsCount;
		     List.app emitMember fields;
		     emitU2i methodsCount;
		     List.app emitMember methods;
		     emitAttrs attrs)
		else classError "emit: too many methods"
	    else classError "emit: too many fields"
	else classError "emit: too many direct superinterfaces"
    end

fun emit out cp decl =
    emitClassFile out (fromClassDecl cp decl)
    
fun scanClassFile src = raise Bytecode.Unimplemented "Classfile.scanClassFile"

fun scan src = toClassDecl(scanClassFile src)
